#ifndef Approximator_H
#define Approximator_H

/// \file Approximator.h
/// \brief Defines the class Approximator

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <cassert>

#include <Random.H>
#include <Geometry.H>

#include <iostream>
#include <fstream>

#ifdef _MSC_VER
#define LONGLONG __int64
#else
#define LONGLONG long long
#endif

/// \brief Common base class for learning/representing a value function over an
/// Euclidean state space, and with a finite number of actions.
class Approximator {
public:
  explicit Approximator(const int dimension, const int actions)
    :m_dimension(dimension), m_actions(actions),m_flops(0) {
      assert(dimension > 0); assert(actions > 0);}

  virtual ~Approximator() {}

  int dimension() const {return m_dimension;}
  int numActions() const {return m_actions;}

  LONGLONG flops() const { return m_flops; }
  LONGLONG  set_flops( unsigned long f ) { m_flops = f; return m_flops; }
  virtual int mem_usage() const { return 0; }

  virtual void learn(RnPoint &point, const int action, const double value) {}
  
  virtual void learn(RnPoint &state, const int action, const double reward,
	            	     RnPoint &next_state, bool goalState) = 0;

  virtual double present(RnPoint &point, const int action) { return 0.0;}

  virtual void runCache() = 0;
  virtual int  numPoints() const = 0;
  virtual void initialize(RealRandom &generator) = 0;

  /// \brief Q-value of a point and an action.
  double q(RnPoint &point, const int action) {
    return present(point, action);}

  /// \brief Value of a point and an action.
  double maxQ(RnPoint &point) {
    int action; double value; greedy(point, value, action); return value;}
  
  /// \brief Returns index of best action for the given point.
  int greedyAction(RnPoint &point) {
    int action; double value; greedy(point, value, action); return action;}

  /// \brief Reads object from stream.
  virtual void read(std::istream &in) {}
  /// \brief Writes object to stream.
  virtual void write(std::ostream &out) const {}

  void read(const char *const filename) {
    assert(filename != NULL); std::ifstream in(filename); read(in);}
  void write(const char *const filename) const {
    assert(filename != NULL); std::ofstream out(filename); write(out);}

protected:
  /// \brief Default kernel to be used for interpolation.
  double kernel(const double distance, const double bandwidth) const {
    m_flops += 4;
    return exp((-4.6 * distance * distance) / (bandwidth * bandwidth));
  }

private:
  const int m_dimension;  ///< Dimensionality of state space
  const int m_actions;    ///< Number of actions.
protected:
  mutable LONGLONG m_flops; ///< No. of floating point operations so far.

private:
  void greedy(RnPoint &point, double &value, int &action) {
    action = 0; value = q(point, 0);
    for(int i = 1; i < m_actions; ++i) {
      const double val = q(point, i);
      if (val > value) {action = i; value = val;}
    }
    m_flops += m_actions;
  }

  // Not allowed to use these
  Approximator(const Approximator &approx);
  Approximator &operator=(const Approximator &approx);
};

#endif
