#ifndef DescriptiveStats_H
#define DescriptiveStats_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <iostream>
#include <cmath>

#include <cfloat>
#include <cstring>
#include <cassert>


/// Collects 1D data and computes a number of simple statistics for the data collected.
class DescriptiveStats {
  friend std::ostream &operator<<(std::ostream &out,
				  const DescriptiveStats &stats) {
    stats.summary(out); return out;}
  
public:
  explicit DescriptiveStats(const char *const description = "")
    :m_sum(0.0), m_sum_squared(0.0), m_min(DBL_MAX), m_max(-DBL_MAX),

     m_data_points(0) {assert(description != NULL);
     strncpy(m_description, description, 40); m_description[40] = '\0';}
  DescriptiveStats(const DescriptiveStats &stats)
    :m_sum(stats.m_sum), m_sum_squared(stats.m_sum_squared),
     m_min(stats.m_min), m_max(stats.m_max),
     m_data_points(stats.m_data_points) {
    strcpy(m_description, stats.m_description);}

  DescriptiveStats &operator=(const DescriptiveStats &stats) {
    m_sum = stats.m_sum; m_sum_squared = stats.m_sum_squared;
    m_min = stats.m_min; m_max = stats.m_max;
    m_data_points = stats.m_data_points;
    strcpy(m_description, stats.m_description); return *this;}

  void add(const double number) {
    ++m_data_points; if (number < m_min) m_min = number;
    if (number > m_max) m_max = number;
    m_sum += number; m_sum_squared += number * number;}
  DescriptiveStats &operator<<(const double number) {
    add(number); return *this;}
  void operator+=(const double number) {add(number);}

  void reset() {m_sum = 0.0; m_sum_squared = 0.0; m_min = DBL_MAX;
  m_max = -DBL_MAX; m_data_points = 0;}

  void summary(std::ostream &out) const;

  int dataPoints() const {return m_data_points;}

  double min() const {assert(m_data_points > 0); return m_min;}
  double max() const {assert(m_data_points > 0); return m_max;}

  double range() const {assert(m_data_points > 0); return m_max - m_min;}

  double sum() const {return m_sum;}
  double sumSquared() const {return m_sum_squared;}
  
  double mean() const {
    assert(m_data_points > 0); return m_sum / (double)m_data_points;}
  double meanSquared() const {
    assert(m_data_points > 0); return m_sum_squared / (double)m_data_points;}
  double variance() const {
    assert(m_data_points > 1); 
    return (m_sum_squared - (m_sum * (m_sum / (double)m_data_points))) /
      ((double)m_data_points - 1.0);}
  double stdDev() const {assert(m_data_points > 1); return sqrt(variance());}

  double largeSampleCI(const double z, double &low, double &high) const {
    assert(m_data_points > 1);

    const double data_points = (double)m_data_points;
    const double delta = (z * sqrt(variance() / data_points));
    low = mean() - delta; high = mean() + delta;
    return delta;}
  double largeSampleCI80(double &low, double &high) const {
    assert(m_data_points > 1); return largeSampleCI(1.28, low, high);}
  double largeSampleCI90(double &low, double &high) const {
    assert(m_data_points > 1); return largeSampleCI(1.645, low, high);}
  double largeSampleCI95(double &low, double &high) const {
    assert(m_data_points > 1); return largeSampleCI(1.96, low, high);}
  double largeSampleCI99(double &low, double &high) const {
    assert(m_data_points > 1); return largeSampleCI(2.5758, low, high);}

private:
  double m_sum;
  double m_sum_squared;

  double m_min;
  double m_max;
  
  int m_data_points;

  char m_description[41];
};


inline void DescriptiveStats::summary(std::ostream &out) const {
  // Do we have a description?
  if (m_description[0] != '\0')
    out << m_description << std::endl;

  out << "  Data points: " << m_data_points << std::endl;
  
  // Enough for mean?
  if (m_data_points > 0) {
    out << "  Mean: " << mean() << std::endl;

    // Enough for variance?
    if (m_data_points > 1)
      out << "  Variance: " << variance() << std::endl << "  Std Dev: " <<
        stdDev() << std::endl;

    out << "  Min: " << m_min << std::endl << "  Max: " << m_max << std::endl;
  }
  
  out << "  Sum: " << m_sum << std::endl;

  // Enough for a confidence interval?
  if (m_data_points > 1) {
    double low, high;
    largeSampleCI95(low, high);
    out << "  Large sample CIs:" << std::endl << "    95%: (" << low <<
      ", " << high << ")" << std::endl;
    largeSampleCI99(low, high);
    out << "    99%: (" << low << ", " << high << ")" << std::endl;
  }
}


#endif
