#ifndef HardStateAggregator_H
#define HardStateAggregator_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <cmath>
#include <Array.H>
#include <Random.H>
#include <debug.h>

/// \brief Q-learning/Q-table representation for continuous state space MDPs
/// employing (hard) state aggregation based on a grid-based discretization 
/// of the state space.
/// 
/// This is probably the oldest and simplest approach to making RL work
/// in continuous spaces.
class HardStateAggregator {
public:
  explicit HardStateAggregator(const int states, const int actions, const double gain,
	 const double discount)
    :m_gain(gain), m_discount(discount), m_data(states, actions) {
    assert(states > 0); assert(actions > 0);
    assert_range_inc(m_discount, 0.0, 1.0);}
  HardStateAggregator(const HardStateAggregator &table)
    :m_gain(table.m_gain), m_discount(table.m_discount),
     m_data(table.m_data) {}

  void initialize(const double number) {m_data = number;}
  void initialize(RealRandom &generator) {
    for(int i = 0; i < numStates(); ++i) for(int j = 0; j < numActions(); ++j)
      m_data(i, j) = generator.number();}

  double q(const int state, const int action) const {
    assert_range(state, 0, numStates()); assert_range(action, 0, numActions());
    return m_data(state, action);}
  double qValue(const int state, const int action) const {
    assert_range(state, 0, numStates()); assert_range(action, 0, numActions());
    return q(state, action);}

  void learn(const int state, const int action, const double reward,
	     const int next_state) {
    assert_range(state, 0, numStates()); assert_range(action, 0, numActions());
    assert_range(next_state, 0, numStates());
    double next_value; greedyAction(next_state, next_value);
    m_data[state][action] = m_data[state][action] * (1.0 - m_gain) +
      m_gain * (reward + m_discount * next_value);}

  void learn(const int state, const int action, const double reward,
	     const int next_state, bool episode_end ) {
    assert_range(state, 0, numStates()); assert_range(action, 0, numActions());
    assert_range(next_state, 0, numStates());
    double next_value = 0;
    if (!episode_end) greedyAction(next_state, next_value);
    m_data[state][action] = m_data[state][action] * (1.0 - m_gain) +
      m_gain * (reward + m_discount * next_value);}

  void learn(const int state, const int action, const double reward,
	     const int next_state, const int next_action) {
    assert_range(state, 0, numStates()); assert_range(action, 0, numActions());
    assert_range(next_state, 0, numStates());
    assert_range(next_action, 0, numActions());
    m_data[state][action] = m_data[state][action] * (1.0 - m_gain) +
      m_gain * (reward + m_discount * qValue(next_state, next_action));}

  int greedyAction(const int state) const { 
    assert_range(state, 0, numStates());
    double dummy; return greedyAction(state, dummy);}

  int greedyAction(const int state, double &value) const {
    assert_range(state, 0, numStates());
    int best_action = 0; value = qValue(state, 0);
    for(int i = 1; i < numActions(); ++i)
      if (qValue(state, i) > value){
	best_action = i; value = qValue(state, i);}
    return best_action;}

  int boltzmanAction(const int state, const double temperature,
		     const double random_number) const;

  int numStates() const {return m_data.rows();}
  int numActions() const {return m_data.columns();}

  double gain() const {return m_gain;}
  void setGain(const double gain) {m_gain = gain;}

  double discount() const {return m_discount;}
  void setDiscount(const double discount) {m_discount = discount;}

private:
  double m_gain;
  double m_discount;

  Array2d <double> m_data;

  // Not allowed to use this
  HardStateAggregator &operator=(const HardStateAggregator &table);
};


inline int HardStateAggregator::boltzmanAction(const int state, const double temperature,
				  const double random_number) const {
  assert_range(state, 0, numStates());
  assert(temperature > 0.0);
  assert_range_inc(random_number, 0, 1.0);
  
  Array <double> boltzman_values(numStates());

  //Calculate and store the Boltzman values
  boltzman_values[0] = exp(qValue(state, 0) / temperature);
  for(int i = 1; i < numStates(); ++i)
    boltzman_values[i] = boltzman_values[i - 1] +
      exp(qValue(state, i) / temperature);

  // Now, select an action
  double number = random_number * boltzman_values[numStates() - 1];
  {
  for(int i = 0; i < numStates(); ++i)
    if (number <= boltzman_values[i])
      return i;
  }

  // Just in case
  return (numStates() - 1);
}


#endif
