#ifndef HillCar_H
#define HillCar_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <cmath>
#include <debug.h>
#include <MDP.H>

/// \brief Car-on-the hill task (dynamics + rewards).
class HillCar : public ContinuousStateMDP, public ContinuousActionMDP,
		public DiscreteActionMDP {
public:
  enum Action {Left = 0, Coast = 1, Right = 2};

  HillCar()
    :ContinuousStateMDP(), ContinuousActionMDP(), DiscreteActionMDP(),
     m_position(M_PI / -6.0), m_velocity(0.0), m_parking_reward(false) {}
  HillCar(const HillCar &car)
    :ContinuousStateMDP(car), ContinuousActionMDP(car), DiscreteActionMDP(car),
     m_position(car.m_position), m_velocity(car.m_velocity),
     m_parking_reward(false) {}

  HillCar &operator=(const HillCar &car) {
    ContinuousStateMDP::operator=(car); ContinuousActionMDP::operator=(car);
    DiscreteActionMDP::operator=(car);
    m_position = car.m_position; m_velocity = car.m_velocity;
    m_parking_reward = car.m_parking_reward; return *this;}

  virtual int stateDimension() const {return 2;}
  virtual int actionDimension() const {return 1;}
  virtual int numActions() const {return 3;}

  double position() const {return m_position;}
  void setPosition(const double position) {
    assert(position >= -1.2); assert(position <= 0.5);
    m_position = position; cropPosition();}

  double velocity() const {return m_velocity;}
  void setVelocity(const double velocity) {
    assert(velocity >= -0.07); assert(velocity <= 0.07);
    m_velocity = velocity; cropVelocity();}

  bool parkingReward() const {return m_parking_reward;}
  void setParkingReward() {m_parking_reward = true;}
  void resetParkingReward() {m_parking_reward = false;}

  virtual void state(double *const state) const {
    assert_pointer(state); state[0] = m_position; state[1] = m_velocity;}
  virtual double state(const int index) const {
    assert_range(index, 0, 2);return (index == 0)?(m_position):(m_velocity);}

  virtual void setState(const double *const state) {
    assert_pointer(state); assert_range_inc(state[0], -1.2, 0.5);
    assert_range_inc(state[1], -0.07, 0.07);
    m_position = state[0]; m_velocity = state[1];}
  virtual void setState(const int index, const double value) {
    assert_range(index, 0, 2);
    if (index == 0) {assert_range_inc(value, -1.2, 0.5); m_position = value;}
    else {assert_range_inc(value, -0.07, 0.07); m_velocity = value;}}
    
  virtual void reset() {m_position = M_PI / -6.0; m_velocity = 0.0;}

  virtual void normalizeState(double *const state) const {
    assert_pointer(state); assert_range_inc(state[0], -1.2, 0.5);
    assert_range_inc(state[1], -0.07, 0.07);
    state[0] = (state[0] + 1.2) / 1.7; state[1] = (state[1] + 0.07) / 0.14;}
  virtual void deNormalizeState(double *const state) const {
    assert_pointer(state); assert_range_inc(state[0], 0.0, 1.0);
    assert_range_inc(state[1], 0.0, 1.0);
    state[0] = (state[0] * 1.7) - 1.2; state[1] = (state[1] * 0.14) - 0.07;}

  virtual bool goalState(const double epsilon = 0.001) const {
    return (0.5 - m_position) < epsilon;}

  virtual double applyAction(const double *const action);
  virtual double applyAction(const int action) {
    assert_range(action, 0, 3);
    if (action == Left) return applyLeftAction();
    else if (action == Right) return applyRightAction();
    else return applyCoastAction();}

  double ApplyAction(const Action action) {return applyAction((int)action);}
  double applyLeftAction() {return ContinuousActionMDP::applyAction(-1.0);}
  double applyCoastAction() {return ContinuousActionMDP::applyAction(0.0);}
  double applyRightAction() {return ContinuousActionMDP::applyAction(1.0);}

  virtual void deNormalizeAction(double *const action) const {
    assert_pointer(action); action[0] = (action[0] * 2.0) - 1.0;}

private:
  double m_position;
  double m_velocity;

  bool m_parking_reward;

  // Simulation parameters
#ifndef _MSC_VER
  static const double s_time_step = 0.001;
  static const double s_magic_gravity_number = 0.0025;
#else
  #define s_time_step 0.001
  #define s_magic_gravity_number 0.0025
#endif

  void cropPosition() {
    if (m_position < -1.2) m_position = -1.2;
    else if (m_position > 0.5) m_position = 0.5;}
  void cropVelocity() {
    if (m_velocity < -0.07) m_velocity = -0.07;
    else if (m_velocity > 0.07) m_velocity = 0.07;}
};

inline double HillCar::applyAction(const double *const action) {
  assert_pointer(action);

  m_velocity += (s_time_step * action[0]) -
    (s_magic_gravity_number * cos(3.0 * m_position));
  cropVelocity();

  m_position += m_velocity;
  cropPosition();

  double reward = 0.0;
  if (goalState()) {
    if (m_parking_reward)
      reward += (1.0 - (m_velocity / 0.07));
    else
      reward += 1.0;
  } else if (m_position <= -1.2) {
    reward -= 1.0; m_velocity = 0.0;
  }

  return reward;
}

#endif
