#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <IFAPPQ.H>

#include <algorithm>
#include <cfloat>

#include <debug.h>

using namespace std;

// going to use a Gaussian kernel for 'interpolation'
#define GAUSSIAN_KERNEL


IFAPPQ::IFAPPQ(const int dimension, const int actions, const double gain,
	       const double discount, const double threshold,
	       const double epsilon, const double count_factor,
	       const double initial, const bool random)
  :Approximator(dimension, actions), m_threshold(threshold), m_gain(gain),
   m_discount(discount), m_epsilon(epsilon),m_count_factor(count_factor) {
  assert(threshold >= 0.0);
  assert(epsilon > 0);

  int cells = 20;
  // Find out how many points we have to allocate
  int num_points = 1;
  for(int i = 0; i < dimension; num_points *= cells, ++i);

  RnPoint point(dimension);
  RnPoint values(actions, initial);
  
  point[0] = 0.5;
  point[1] = 0.5;

  // Put in in initial seed point
  ApproximatorPoint ap(point, values, 0.0, DBL_MAX);
  m_data.push_back(ap);
}

void IFAPPQ::initialize(RealRandom &generator) {
  for(vector<ApproximatorPoint>::iterator it = m_data.begin();
      it != m_data.end(); ++it)
    for(int i = 0; i < numActions(); ++i)
      it->values[i] = generator;
}



void IFAPPQ::learn(RnPoint &point, const int action, const double value) {
  assert_range(action, 0, numActions());

  // Find the closest point
  vector<ApproximatorPoint>::iterator closest = m_data.begin();
  double closest_dist = point.distance(closest->point);
  m_flops += dimension()*3+2;
  for(vector<ApproximatorPoint>::iterator it = m_data.begin();
      it != m_data.end(); ++it) {
    const double dist = point.distance(it->point);
    if (dist < closest_dist) {
      closest_dist = dist;
      closest = it;
    }
  }
  m_flops += m_data.size()*(dimension()*3+2);

  // Do we need to add a new point?
  m_flops += 1;
  if (closest_dist > m_threshold) {
    ApproximatorPoint ap(*closest);
    ap.point = point;
    ap.distance = 0.0;
    for (int i=numActions()-1; i>=0; --i)
    {
      ap.values[i] = q(point,i);
    }
    m_data.push_back(ap);

    closest = m_data.end() - 1;

    // Calculate the new distances between the points
    calculateDistances();
  }

  {
    for(vector<ApproximatorPoint>::iterator it = m_data.begin();
        it != m_data.end(); ++it) {
      const double k = kernel(point.distance(it->point), it->bandwidth);

      if (k > m_epsilon) {
        it->values[action] +=
	  (m_gain / (1.0 + double(it->count)/m_count_factor)) * k *
	  (value - it->values[action]);
        ++(it->count);
        m_flops += 6;
      }
    }
  }
}


void IFAPPQ::runCache() {
  vector<Experience>::reverse_iterator it;
  for(it = m_cache.rbegin(); it != m_cache.rend(); ++it) {
    actualLearn(it->state, it->action, it->reward, it->next_state,it->goalState );
  }
  m_cache.clear();
}



double IFAPPQ::present(RnPoint &point, const int action) {
  assert_range(action, 0, numActions());

  // Work out the value
  double sum = 0.0;
  double kernel_sum = 0.0;
  for(vector<ApproximatorPoint>::iterator it = m_data.begin();
      it != m_data.end(); ++it) {
    sum += it->values[action] * kernel(point.distance(it->point),
				       it->bandwidth);
    kernel_sum += kernel(point.distance(it->point), it->bandwidth);
    m_flops += 3;
  }

#ifndef GAUSSIAN_KERNEL
  if (kernel_sum == 0.0) {
    sum = 0.0;
    for(vector<ApproximatorPoint>::iterator it = m_data.begin();
	      it != m_data.end(); ++it)
          sum += it->values[action];
    sum /= (double)(m_data.size());
    m_flops += m_data.size()+1;
  } else {
#endif
    sum /= kernel_sum;
    m_flops += 1;
#ifndef GAUSSIAN_KERNEL
  }
#endif

  return sum;
}



void IFAPPQ::read(istream &in) {
  // Empty the list of points
  while (!m_data.empty())
    m_data.pop_back();

  RnPoint point(dimension());
  RnPoint values(numActions());

  ApproximatorPoint ap(point, values);

  in >> ap;
  while (!in.eof()) {
    m_data.push_back(ap);
    
    in >> ap;
  }
}



void IFAPPQ::write(ostream &out) const {
  for(vector<ApproximatorPoint>::const_iterator it = m_data.begin();
      it != m_data.end(); ++it)
    out << *it << endl;
}



void IFAPPQ::calculateDistances() {
  for(vector<ApproximatorPoint>::iterator it = m_data.begin();
      it != m_data.end(); ++it) {
    for(vector<ApproximatorPoint>::iterator it2 = m_data.begin();
        it2 != m_data.end(); ++it2) {
      if (it != it2) {
        const double dist = (it->point).distance(it2->point);
        if (dist < it->bandwidth) {
          it->bandwidth = dist;
        }
        m_flops += 3*dimension()+2;
      }
    }
  }
}
