#ifndef _RLFAPP_KBRL_H
#define _RLFAPP_KBRL_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <vector>
#include <iostream>
#include <Random.H>
#include <Geometry.H>
#include <debug.h>
#include <Approximator.H>
#include <Experience.H>

using namespace std;

#ifdef _MSC_VER
#define ULONGLONG unsigned __int64
#else
#define ULONGLONG unsigned long long
#endif

/// Helper class for solving a finite MDP using value iteration
class FiniteMDP 
{
public:
  FiniteMDP();
  explicit FiniteMDP( int nStates, int nActions, double discount );

  void reset( int nStates, int nActions, double discount );

  void set_prob( int start, int action, int end, double prob )
  {
    assert_range( start,  0, m_nStates );
    assert_range( end,    0, m_nStates );
    assert_range( action, 0, m_nActions);
    if (prob>0.0)
      m_prob[start][action].push_back( make_pair(end,prob) );
  }

  void set_reward( int start, int action, double reward )
  {
    assert_range( start,  0, m_nStates );
    assert_range( action, 0, m_nActions);
    m_reward[start][action] = reward;
  }

  ULONGLONG flops() const { return m_flops; }

  /// Use value iteration to solve the finite MDP
  double solve( double epsilon, int iter_max );

  double value( int state ) const { return m_value[state]; }
  double reward( int start, int action ) const
  { 
    assert_range( start,  0, m_nStates );
    assert_range( action, 0, m_nActions);
    return m_reward[start][action]; 
  }

  double prob( int start, int action, int end ) const
  {
    assert_range( start,  0, m_nStates );
    assert_range( end,    0, m_nStates );
    assert_range( action, 0, m_nActions);
    for (int k=m_prob[start][action].size()-1; k>=0; --k)
    {
      if (m_prob[start][action][k].first==end)
      {
        return m_prob[start][action][k].second;
      }
    }
    return 0.0;
  }

  int mem_usage() const;

protected:
  double full_backup();
  void checkTransProbs() const;

protected:
  // assumes that many transitions are zero!
  vector< vector< vector< pair<int,double> > > > m_prob;
  vector< vector< double > > m_reward;
  int m_nStates;
  int m_nActions;
  vector< double > m_value;
  vector< double > m_next_value;
  double m_discount;
  ULONGLONG m_flops;
};

/// \brief Kernel-based reinforcement learning
/// 
/// The method works as follows: non-parametric (kernel based)
/// estimates of the transition kernel and the rewards are 
/// developed. Then Rust's idea is used to solve the Bellman equation.
/// For more information see Ormoneit and Sen, 
/// Kernel-Based Reinforcement Learning, Volume 49, Issue 2, Nov 2002 
/// Pages: 161-178, also
/// http://citeseer.ist.psu.edu/ormoneit99kernelbased.html
class KBRL : public Approximator
{
public:
  explicit KBRL(const int dimension, const int actions,
		   const double discount,
		   const double epsilon = 0.01,
       const int iterations = 100,
		   const double initial = 0.0
      );

  void initialize(RealRandom &generator) {}

  void learn(RnPoint &state, const int action, const double reward,
	     RnPoint &next_state, bool goalState) {
    m_data[action].push_back(Experience(state, action, reward, next_state,
					goalState));
    ++m_numPoints;
    m_solved = false;
  }
  
  void learn(const RnPoint &point, const int action, const double value) {
    assert(false); // no one is expected to use this function!
  }

  double present(RnPoint &point, const int action) {
    assert(false); return 0.0;} // not expected to call this

  void runCache();

  double present(const RnPoint &point, const int action);

  int numPoints() const { return m_numPoints; }

  void read(std::istream &in);
  void write(std::ostream &out) const;
  int mem_usage() const;

protected:
  void computeStateStartIndex();
  void computeBandwidth();
  int state_index( int l, int action ) const {
    return m_stateStartIndex[action]+l;
  }
  double pfloor( double p ) const {
    return (p<m_floor?0.0:p);
  }
  double kernel(const RnPoint& p1, const RnPoint& p2) const {
    return Approximator::kernel( p1.distance(p2), m_bandwidth );
    m_flops += 3*dimension()+1;
  }

private:
  int                           m_numPoints;
  vector<vector<Experience> >   m_data;
  double                        m_discount;
  double                        m_epsilon;
  int                           m_iterations;
  double                        m_floor;
  FiniteMDP                     m_finiteMDP;
  bool                          m_solved;
  vector<int>                   m_stateStartIndex; // one index for each action
  double                        m_bandwidth;
};

#endif //_RLFAPP_KBRL_H
