#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <KBRL.H>
#include <cmath>
#include <utility>
#include <limits>


#ifdef _MSC_VER
namespace std
{
  template <class T>
  inline T abs( T a ) { return (a>=0?a:0); }
}
#endif


FiniteMDP::FiniteMDP() 
: m_nStates( 0 )
, m_nActions( 0)
, m_discount( 0.9 )
, m_flops( 0 )
{
}

FiniteMDP::FiniteMDP( int nStates, int nActions, double discount )
: m_prob( nStates, 
          vector< vector< pair<int,double> > >( nActions
            , vector< pair<int,double> >() ) )
, m_reward( nStates, vector<double>( nActions, -std::numeric_limits<double>::infinity() ) )
, m_nStates( nStates )
, m_nActions( nActions )
, m_value( nStates, 0.0 )
, m_discount( discount )
{
  assert_range( discount, 0.0, 1.0 );
  for (int i=0; i<nStates; ++i)
  {
    for (int j=0; j<nActions; ++j)
    {
      set_prob( i,j,i, 1.0 );
    }
  }
}

void FiniteMDP::reset( int nStates, int nActions, double discount )
{
  m_prob.clear();
  m_reward.clear();
  m_value.clear();
  m_next_value.clear();
  m_prob.resize( nStates, 
          vector< vector< pair<int,double> > >( nActions
            , vector< pair<int,double> >() ) );
  m_reward.resize( nStates, vector<double>( nActions, -std::numeric_limits<double>::infinity() ) );
  m_nStates = ( nStates );
  m_nActions = ( nActions );
  m_value.resize( nStates, 0.0 );
  m_discount = ( discount );
  //m_flops = 0;
}

int FiniteMDP::mem_usage() const
{
  int sz_n = 0;
  for (int i=0; i<m_nStates; ++i)
  {
    for (int j=0; j<m_nActions; ++j)
    {
      sz_n += m_prob[i][j].size();
    }
  }
  sz_n += m_nStates * 2; // rewards; 2 actions
  sz_n += m_nStates * 2; // 2 value tables
  return sz_n;
}


void FiniteMDP::checkTransProbs() const
{
  int sz_n = 0;
  int sz_d = 0;
  for (int i=0; i<m_nStates; ++i)
  {
    for (int j=0; j<m_nActions; ++j)
    {
      double v = 0.0;
      for (int k=m_prob[i][j].size()-1; k>=0; --k)
      {
        v += m_prob[i][j][k].second;
      }
      if(m_prob[i][j].size()>0)
      {
        ++sz_d;
        sz_n += m_prob[i][j].size();
        if (v<0.999 || v>1.001)
        {
          std::cout << "Problem with (state,action) = (" << i << ", " << j << ")\n";
          std::cout << "Sum of transition probabilities is " << v << "\n";
        }
      }
    }
  }
  std::cout << "Total number of connections: " << sz_n << "\n";
  std::cout << "Total number of states-action pairs: " << sz_d << "\n";
  std::cout << "Average perplexity: " << float(sz_n)/sz_d << "\n";
  std::cout << "Check of transition probabilities is ready.\n";

  // dump rewards
  if (0)
  {
    std::ofstream os("reward_table");
    double rmax = 0.0;
    for (int i=0; i<m_nStates; ++i)
    {
      for (int j=0; j<m_nActions; ++j)
      {
        os << i << "\t" << j << "\t" << reward(i,j) << "\n";
        rmax = (reward(i,j)>rmax?reward(i,j):rmax);
      }
    }
    std::cout << "Reward table dump ready.\n";
    std::cout << "Rmax = " << rmax << std::endl;
  }
}

double FiniteMDP::solve( double epsilon, int iter_max )
{
  checkTransProbs();
  // use value iteration to solve the MDP
  double e = 2*epsilon;
  for (int i=0; i<iter_max && e>epsilon; ++i)
  {
    e = full_backup();
  }
  return e;
}

double FiniteMDP::full_backup()
{
  double l1_error = 0.0;
  double max_error = 0.0;
  double e;
  m_next_value.resize( m_nStates );
  for (int i=0; i<m_nStates; ++i)
  {
    double best_value = -numeric_limits<double>::infinity();
    for (int j=0; j<m_nActions; ++j)
    {
      double v = 0.0;
      for (int k=m_prob[i][j].size()-1; k>=0; --k)
      {
        v += m_prob[i][j][k].second * value(m_prob[i][j][k].first);
      }
      v *= m_discount;
      v += reward(i,j);
      best_value = (v>best_value?v:best_value);
      m_flops += m_prob[i][j].size() + 3;
    }
    m_next_value[i] = best_value;
    e = std::abs(m_next_value[i]-m_value[i]);
    l1_error += e;
    max_error = (e>max_error?e:max_error);
  }
  l1_error /= m_nStates;
  m_value.swap(m_next_value);

  return max_error;
}


KBRL::KBRL(const int dimension, 
           const int actions,
	   const double discount,
	   const double epsilon, 
           const int iterations,
           const double floor   
          )
: Approximator(dimension, actions)
, m_numPoints(0)
, m_data(actions)
, m_discount(discount)
, m_epsilon(epsilon)
, m_iterations( iterations )
, m_floor(floor)
, m_solved(false)
, m_bandwidth(0.0)
{
}

void KBRL::computeBandwidth()
{
  m_bandwidth = 1.0/sqrt((double)m_numPoints)* (1.0+ log(1.0+m_numPoints/750.0));
}

void KBRL::computeStateStartIndex()
{
  m_stateStartIndex.resize( numActions(), 0 );
  int m = 0;
  for (int i=0; i<numActions(); ++i)
  {
    m_stateStartIndex[i] = m;
    m+=m_data[i].size();
  }
}

void KBRL::runCache()
{
  // compute finiteMDP
  m_finiteMDP.reset( m_numPoints, numActions(), m_discount );

  computeStateStartIndex();
  computeBandwidth();

  for (int b=0; b<numActions(); ++b)
  {
    int e_b_sz = m_data[b].size();
    for (int u=0; u<e_b_sz; ++u)            // iterate over (u,b)
    {
      int start_index = state_index(u,b);
      for (int a=0; a<numActions(); ++a)    // given action index a
      {
        int e_a_sz = m_data[a].size();      // compute norm factor for (u,b),a
        double ksum_a_u_b = 0;
        for (int z=0; z<e_a_sz; ++z)
        {
          ksum_a_u_b += kernel( m_data[a][z].state, m_data[b][u].next_state );
        }
        if (ksum_a_u_b<1E-10)
        {
          ksum_a_u_b = 1.0;
        }
        m_flops += e_a_sz*(dimension()*3+5);
        double r = 0;
        for (int l=0; l<e_a_sz; ++l)        // iterate over (l,a)
        {
          int end_index = state_index(l,a);
          double k = kernel( m_data[a][l].state, m_data[b][u].next_state );
          k /= ksum_a_u_b;
          r += k * m_data[a][l].reward;
          k = pfloor( k );
          if (m_data[a][l].goalState) k = 0.0;
          m_finiteMDP.set_prob( start_index, a, end_index, k );
        }
        m_flops += e_a_sz*3;
        m_finiteMDP.set_reward( start_index, a, r );
      }
    }
  }

  // compute the value function of the finite MDP
  m_finiteMDP.solve(m_epsilon,m_iterations);
  m_flops += m_finiteMDP.flops();
  m_solved = true;
}

double KBRL::present(const RnPoint& point, const int action) 
{
  assert( m_solved );
  double v = 0;
  int e_action_sz = m_data[action].size();
  int l;
  double ksum = 0;
  for (l=0; l<e_action_sz; ++l)
  {
    ksum += kernel( m_data[action][l].state, point );
  }
  m_flops += (dimension()*3+5)*e_action_sz;
  if (ksum<1E-10)
  {
    return 0.0;
  }
  for (l=0; l<e_action_sz; ++l)
  {
    v += kernel( m_data[action][l].state, point ) * 
      (m_data[action][l].reward+
       (m_data[action][l].goalState?0.0:
        m_discount*m_finiteMDP.value( state_index(l,action) )));
  }
  v /= ksum;
  m_flops += 2*e_action_sz+1;
  return v;
}

void KBRL::read(std::istream &in)
{
  int dim;
  in >> dim;
  assert( dim==dimension() );
  RnPoint p1(dim),p2(dim);
  int nActions;
  in >> nActions;
  assert( nActions == numActions() );
  for (int i=0; i<numActions(); ++i)
  {
    int j, sz;
    in >> j >> sz;
    assert( i==j );
    m_data[i].resize( sz, Experience(p1,0,0,p2,false));
    for (j=0; j<sz; ++j)
    {
      in >> m_data[i][j];
    }
  }
}

void KBRL::write(std::ostream &out) const
{
  out << dimension() << "\n";
  out << numActions() << "\n";
  for (int i=0; i<numActions(); ++i)
  {
    out << i << "\t" << m_data[i].size() << "\n";
    for (size_t j=0; j<m_data[i].size(); ++j)
    {
      out << "\t" << m_data[i][j] << "\n";
    }
  }
}

int KBRL::mem_usage() const
{
  int sz_n = 0;
  for (int i=0; i<numActions(); ++i)
  {
    sz_n += m_data[i].size()*(dimension()*2+2);
  }
  sz_n += m_finiteMDP.mem_usage();
  return sz_n;
}
