#ifndef MDP_H
#define MDP_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <iostream>

#include <Geometry.H>
#include <Random.H>

#include <debug.h>
#include <vector>

using namespace std;

//----------------------------------------------------------------------
// States
//----------------------------------------------------------------------

/// \brief Interface for continuous state MDPs.
class ContinuousStateMDP {
  friend std::ostream &operator<<(std::ostream &out,
				  const ContinuousStateMDP &mdp) {
    out << mdp.state(0);
    for(int i = 1; i < mdp.stateDimension(); ++i) out << ' ' << mdp.state(i);
    return out;}
  friend std::istream &operator>>(std::istream &in, ContinuousStateMDP &mdp) {
    double tmp; for(int i = 0; i < mdp.stateDimension(); ++i) {
      in >> tmp; mdp.setState(i, tmp);} return in;}

public:
  ContinuousStateMDP() {}
  ContinuousStateMDP(const ContinuousStateMDP &mdp) {}
  virtual ~ContinuousStateMDP() {}

  ContinuousStateMDP &operator=(const ContinuousStateMDP &mdp) {return *this;}
  double operator[](const int index) const {
    assert_range(index, 0, stateDimension()); return state(index);}

  virtual int stateDimension() const = 0;
  
  virtual void state(double *const state) const = 0;
  virtual double state(const int index) const = 0;

  virtual void setState(const double *const state) = 0;
  virtual void setState(const int index, const double value) = 0;
  virtual void reset() = 0;

  virtual void normalizeState(double *const state) const = 0;
  virtual void deNormalizeState(double *const state) const = 0;

  virtual bool goalState(const double epsilon) const = 0;

  void normalizedState(double *const current_state) const {
    assert_pointer(current_state); state(current_state);
    normalizeState(current_state);}
  double normalizedState(const int index) const {
    assert_range(index, 0, stateDimension()); 
    vector<double> tmp(stateDimension());
    double* ptmp = &tmp[0];
    state(ptmp); normalizeState(ptmp); return tmp[index];
  }
  virtual void setNormalizedState(const double *const state) {
    assert(state != NULL); 
    vector<double> tmp(stateDimension());
    double* ptmp = &tmp[0];
    memcpy(ptmp, state, stateDimension() * sizeof(double));
    deNormalizeState(ptmp); setState(&tmp[0]);}
  // Ouch!  This is ugly.
  virtual void setNormalizedState(const int index, const double value) {
    assert_range(index, 0, stateDimension()); 
    vector<double> tmp(stateDimension());
    double* ptmp = &tmp[0];
    normalizedState(ptmp); tmp[index] = value;
    deNormalizeState(ptmp); setState(index, tmp[index]);}

  void state(RnPoint &point) const {
    assert(point.dimension() == stateDimension()); state((double *)point);}
  void normalizedState(RnPoint &point) const {
    assert(point.dimension() == stateDimension());
    normalizedState((double *)point);}

  void setState(RnPoint &point) {
    assert(point.dimension() == stateDimension());
    setState((const double *const )point);}
  void setNormalizedState(RnPoint &point) {
    assert(point.dimension() == stateDimension());
    setNormalizedState((const double *const )point);}

  void setRandomState() {
#ifdef _MSC_VER
    static UniformRandomSimple number; 
    static callcount = 0;
    if (callcount==0)
    {
      number.randomize();
      ++callcount;
    }
#else
    UniformRandomSimple number; number.randomize();
#endif
    for(int i = 0; i < stateDimension(); ++i) setNormalizedState(i, number);}

private:
};

/// \brief Interface for MDPs with discrete state spaces.
class DiscreteStateMDP {
  friend std::ostream &operator<<(std::ostream &out,
				  const DiscreteStateMDP &mdp) {
    out << mdp.state(); return out;}
  friend std::istream &operator>>(std::istream &in, DiscreteStateMDP &mdp) {
    int tmp; in >> tmp; mdp.setState(tmp); return in;}
				  
public:
  DiscreteStateMDP() {}
  DiscreteStateMDP(const DiscreteStateMDP &mdp) {}
  virtual ~DiscreteStateMDP() {}

  DiscreteStateMDP &operator=(const DiscreteStateMDP &mdp) {return *this;}

  virtual int numStates() const = 0;
  
  virtual int state() const = 0;
  virtual void setState(const int new_state) = 0;
  virtual void reset() = 0;

  virtual bool goalState() const = 0;

  void setRandomState() {
    DiscreteUniformRandom number(numStates()); number.randomize();
    setState(number);}
  
private:
};

//----------------------------------------------------------------------
// Actions
//----------------------------------------------------------------------

/// Interface for MDPs with continuous action space.
class ContinuousActionMDP {
public:
  ContinuousActionMDP() {}
  ContinuousActionMDP(const ContinuousActionMDP &mdp) {}
  virtual ~ContinuousActionMDP() {}

  ContinuousActionMDP &operator=(const ContinuousActionMDP &mdp) {
    return *this;}

  virtual int actionDimension() const = 0;

  virtual double applyAction(const double *const action) = 0;

  virtual void deNormalizeAction(double *const action) const = 0;

  double applyAction(const double action) {
    assert(actionDimension() == 1); return applyAction(&action);}
  double applyNormalizedAction(const double *const action) {
    assert(action != NULL); 
    vector<double> tmp(actionDimension());
    double* ptmp = &tmp[0];
    memcpy(ptmp, action, actionDimension() * sizeof(double));
    deNormalizeAction(ptmp); return applyAction(ptmp);}
  double applyNormalizedAction(const double action) {
    assert(actionDimension() == 1); 
    double tmp = action; deNormalizeAction(&tmp); return applyAction(&tmp);}

  double applyAction(const RnPoint &action) {
    assert(action.dimension() == actionDimension());
    return applyAction((const double *const)action);}

private:
};

/// Interface for MDPs with discrete action sets.
class DiscreteActionMDP {
public:
  DiscreteActionMDP() {}
  DiscreteActionMDP(const DiscreteActionMDP &mdp) {}
  virtual ~DiscreteActionMDP() {}

  DiscreteActionMDP &operator=(const DiscreteActionMDP &mdp) {
    return *this;}

  virtual int numActions() const = 0;

  virtual double applyAction(const int action) = 0;

private:
};

template <class T>
T mmin( T a, T b )
{
  return (a<b?a:b);
}

//----------------------------------------------------------------------
// Helpers
//----------------------------------------------------------------------

/// \brief A helper class to discretize a ContinuousStateMDP
class MDPStateDiscretizer : public DiscreteStateMDP {
public:
  MDPStateDiscretizer(ContinuousStateMDP &mdp, const int number)
    :DiscreteStateMDP(), m_mdp(mdp), m_number(number), m_states(1) {
    assert(number > 0);
    for(int i = 0; i < mdp.stateDimension(); ++i) m_states *= number;}

  virtual int numStates() const {return m_states;}
  
  virtual int state() const {
    int discrete_state = 0;
    const double size = 1.0 / (double)m_number; int mult = 1;
    for(int i = m_mdp.stateDimension() - 1; i >=0 ; --i, mult *= m_number)
        discrete_state += mult * mmin(m_number-1,(int)(m_mdp.normalizedState(i) / size)); //bugfix!
    //if (discrete_state == m_states) discrete_state = m_states - 1;
    assert_range(discrete_state, 0, m_states);
    return discrete_state;}

  virtual void setState(const int new_state) {
//#warning Have to implement this
  }

  virtual bool goalState() const {return false;}

  virtual void reset() {m_mdp.reset();}

private:
  ContinuousStateMDP &m_mdp;

  const int m_number;
  int m_states;

  // Not allowed to use these
  MDPStateDiscretizer(const MDPStateDiscretizer &d);
  MDPStateDiscretizer &operator=(const MDPStateDiscretizer &d);
};

/// \brief A helper class to discretize a ContinuousActionMDP
class MDPActionDiscretizer : public DiscreteActionMDP {
public:
  MDPActionDiscretizer(ContinuousActionMDP &mdp, const int number)
    :DiscreteActionMDP(), m_mdp(mdp), m_number(number),
     m_increment(1.0 / (double)(number - 1)) {
    assert(number > 0); assert(mdp.actionDimension() == 1);}
  virtual ~MDPActionDiscretizer() {}

  virtual int numActions() const {return m_number;}

  virtual double applyAction(const int action) {
    assert_range(action, 0, m_number);
    return m_mdp.applyAction((double)action * m_increment);}

private:
  ContinuousActionMDP &m_mdp;

  const int m_number;
  const double m_increment;

  // Not allowed to use these
  MDPActionDiscretizer(const MDPActionDiscretizer &d);
  MDPActionDiscretizer &operator=(const MDPActionDiscretizer &d);
};


#endif


