#ifndef Random_H
#define Random_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <cmath>
//#include <unistd.h>
//#include <cfile>
#include <cstdio>
#include <fcntl.h>

#ifdef _MSC_VER
#include <ctime>
#include <MathConsts.h>
#else
#include <unistd.h>
#endif// _MSC_VER

//! Various random number generators.
/*! \file Random.H
  This file includes the class definitions for a variety of real- and
  integer-valued random number generators.
  \author Bill Smart
  \ingroup gen_math_group math_group
  \note Some of the random number generators use functions from the
        standard math library.  Be sure to include the math library at
        compile time.
*/


//! Base class for random number generators.
/*! This class is the base for all of the (pseudo) random number generators.
  It is based closely on MersenneTwister.h, v0.6, 14 February 2001,
  written by Richard J. Wagner.  It has been modified and
  substantially trimmed down.
  \author Richard J. Wagner (original code)
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class MTRand {
public:
  /*! This allocates the class and sets the random number seed.
    \param oneSeed Random number seed.  Do \e not set this to zero.
   */
  explicit MTRand(const unsigned long &oneSeed) {randomize(oneSeed);}

  //! \name Randomizing the random variable
  //@{
  //! Set the random number seed using a number generated from \c /dev/urandom
  void randomize();
  
  //! Set the random number seed
  /*! \param seed Random number seed. */
  void randomize(unsigned long seed);
  //@}

  //! \name Moments of the random variable
  //@{
  /*! \return The mean of the random variable. */
  virtual double mean() const = 0;
  /*! \return The variance of the random variable. */
  virtual double variance() const = 0;
  /*! \return The standard deviation of the random variable. */
  virtual double stdDev() const = 0;
  //@}

private:
  typedef unsigned long uint32;
  
protected:
  /*! Basic double precision random number.
    \return Random number in the range 0 to 1. */
  double rand() {return double(randInt()) * 2.3283064370807974e-10;}

  /*! Basic long integer random number.
    \return Random number in the range 0 to \c ULONG_MAX. */
  uint32 randInt();
  /*! Basic bounded long integer random number.
    \param n Largest random number required.
    \return Random number in the range 0 to <i>n</i>. */
  uint32 randInt(const uint32 &n);

private:
  enum {N = 624, M = 397, MAGIC = 0x9908b0dfU};
  
  uint32 state[N];
  uint32 *pNext;
  int left;
  
  void reload();

  uint32 hiBit(const uint32 &u) const {return u & 0x80000000U;}
  uint32 loBit(const uint32 &u) const {return u & 0x00000001U;}
  uint32 loBits(const uint32 &u) const {return u & 0x7fffffffU;}
  uint32 mixBits(const uint32 &u, const uint32 &v) const {
    return hiBit(u) | loBits(v);}
  uint32 twist(const uint32 &m, const uint32 &s0, const uint32 &s1) const {
    return m ^ (mixBits(s0,s1)>>1) ^ (loBit(s1) ? MAGIC : 0U);}

  // Not allowed to use these
  MTRand(const MTRand &mtrand);
  MTRand &operator=(const MTRand &mtrand);
};


//! Base class for real-valued random number generators.
/*! This class is the base for all of the real-valued random number
  generators.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class RealRandom : public MTRand {
public:
  /*! Allocates the class instance and sets the random number seed.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit RealRandom(const unsigned long seed) :MTRand(seed) {}

  //! \name Generate samples
  //@{
  //! Output stream operator.
  /*! This generates a random number and loads it into a double precision
    floating point variable.
    \return A reference to the class instance, used for chaining multiple
    uses of the operator together.
    \param random_number The variable to be assigned to.
  */
  RealRandom &operator>>(double &random_number) {
    random_number = number(); return *this;}
  //! Function operator.
  /*! This allows an instance of the class to look like a function call.
    \return A sample from the random variable.
  */
  double operator()() {return number();}
  //! Double precision cast operator.
  /*! This allows an instance of the class to be treated as a double
    precision floating point number.  Every access of this number will
    return a new sample from the random variable.
    \return A sample from the random variable.
  */
  operator double() {return number();}

  //! Generate a sample from the random variable.
  /*! This function generates a sample from the random variable represented
    by the class instance.
    \return A sample from the random variable.
  */
  virtual double number() = 0;
  //@}

  //! \name Moments of the random variable
  //@{
  virtual double mean() const = 0;
  virtual double variance() const = 0;
  virtual double stdDev() const = 0;
  //@}
};


//! Base class for integer-valued random number generators.
/*! This class is the base for all of the integer-valued random number
  generators.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class IntegerRandom : public MTRand {
public:
  /*! Allocates the class instance and sets the random number seed.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit IntegerRandom(const unsigned long seed) :MTRand(seed) {}

  //! \name Generate samples
  //@{
  //! Output stream operator.
  /*! This generates a random number and loads it into a integer variable.
    \return A reference to the class instance, used for chaining multiple
    uses of the operator together.
    \param random_number The variable to be assigned to.
  */
  IntegerRandom &operator>>(int &random_number) {
    random_number = number(); return *this;}
  //! Function operator.
  /*! This allows an instance of the class to look like a function call.
    \return A sample from the random variable.
  */
  int operator()() {return number();}
  //! Integer cast operator.
  /*! This allows an instance of the class to be treated as an integer
    number.  Every access of this number will return a new sample from
    the random variable.
    \return A sample from the random variable.
  */
  operator int() {return number();}

  //! Generate a sample from the random variable.
  /*! This function generates a sample from the random variable represented
    by the class instance.
    \return A sample from the random variable.
  */
  virtual int number() = 0;
  //@}

  //! \name Moments of the random variable
  //@{
  virtual double mean() const = 0;
  virtual double variance() const = 0;
  virtual double stdDev() const = 0;
  //@}
};


//! Basic uniform random number generator.
/*! This class generates real-valued random numbers uniformly distributed
  between [0.0, 1.0).  This is a replacement for the random number function
  drand48() from stdlib.h.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class UniformRandomSimple : public RealRandom {
public:
  /*! Alocates an instance of the class and sets the random number seed.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit UniformRandomSimple(const unsigned long seed = 1UL) :RealRandom(seed) {}

  //! \name Generate samples
  //@{
  double number() {return MTRand::rand();}
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return 0.5;}
  double variance() const {return (1.0 / 12.0);}
  double stdDev() const {return sqrt(1.0 / 12.0);}
  //@}
};


//! Unform random number generator.
/*! This class generates real-valued random numbers uniformly distributed
  over a given range, [low, high).
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class UniformRandom : public RealRandom {
public:
  /*! Alocates an instance of the class, sets the high and low limits
    for the numbers to be generated, and sets the random number seed.
    \param low Low value for the generated numbers.
    \param high High value for the generated numbers.
    \param seed Random number seed.  Do \e not set this to zero.
    \note This constructor can be used in three different ways.  The
    obvious one is \code 
    UniformRandom gen(low, high); \endcode which generates numbers in
    the range [low, high).  It can also be used with a single parameter
    \code 
    UniformRandom gen(high); \endcode to generate numbers from the range
    [0, high).  The final way is to use all three of the default values
    \code 
    UniformRandom gen; \endcode which generates numbers in the range
    [0.0, 1.0). */
  explicit UniformRandom(const double low = 1.0, const double high = 0.0,
		const unsigned long seed = 1UL)
    :RealRandom(seed), m_low((low < high)?low:high),
     m_range((low < high)?(high - low):(low - high)) {}

  //! \name Generate samples
  //@{
  double number() {return (MTRand::rand() * m_range) + m_low;}
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return ((m_range / 2.0) + m_low);}
  double variance() const {return ((m_range * m_range) / 12.0);}
  double stdDev() const {return sqrt(variance());}
  //@}

  //! \name Parameters of the random variable
  //@{
  /*! \return The low value of the range of numbers that can be generated. */
  double lowValue() const {return m_low;}
  /*! \return The high value of the range of numbers that can be generated. */
  double highValue() const {return (m_low + m_range);}
  /*! \return The range of the numbers that can be generated. */
  double range() const {return m_range;}
  //@}

  //! \name Change parameters of the random variable
  //@{
  /*! Set the low value of the random variable.
    \param low New low value. */
  void setLowValue(const double low) {m_low = low;}
  /*! Set the high value of the random variable.
    \param high New high value. */
  void setHighValue(const double high) {m_range = high - m_low;}  
  /*! Set the both the low and high values of the random variable.
    \param low Low high value.
    \param high New high value. */
  void setValues(const double low, const double high) {
    setLowValue(low); setHighValue(high);}
  //@}

private:
  double m_low;
  double m_range;
};


//! Normal random number generator.
/*! This class generates random numbers from a Gaussian distribution.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
  \note This class uses functions from the standard math library.  Be
        sure to include it on the compile line.
*/
class NormalRandom : public RealRandom {
public:
  /*! This constructor sets the mean and the variance of the random variable,
    and also the random number seed.
    \param mean The mean of the random variable.
    \param variance The variance of the random variable.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit NormalRandom(const double mean = 0.0, const double variance = 1.0,
	       const unsigned long seed = 1UL)
    :RealRandom(seed), m_mean(mean), m_std_dev(sqrt(fabs(variance))) {}
  
  //! \name Generate samples
  //@{
  double number();
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return m_mean;}
  double variance() const {return (m_std_dev * m_std_dev);}
  double stdDev() const {return m_std_dev;}
  //@}

  //! \name Change parameters of the random variable
  //@{
  /*! Set the mean of the random variable.
    \param mean The new mean of the random variable. */
  void setMean(const double mean) {m_mean = mean;}
  /*! Set the variance of the random variable.
    \param variance The new variance of the random variable. */
  void setVariance(const double variance) {m_std_dev = sqrt(fabs(variance));}
  /*! Set the standard deviation of the random variable.
    \param std_dev The new standard deviation of the random variable. */
  void setStdDev(const double std_dev) {m_std_dev = fabs(std_dev);}
  /*! Set the mean and variance of the random variable.
    \param mean The new mean of the random variable.
    \param variance The new variance of the random variable. */
  void setParameters(const double mean, const double variance) {
      setMean(mean); setVariance(variance);}
  //@}

private:
  double m_mean;
  double m_std_dev;
};


//! Exponential random number generator.
/*! This class generates random numbers from an exponential distribution.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class ExponentialRandom : public RealRandom {
public:
  /*! This constructor sets the parameter of the random variable and the
    random number seed.
    \param lambda Parameter of the random variable.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit ExponentialRandom(const double lambda = 1.0, const unsigned long seed = 1UL)
    :RealRandom(seed), m_lambda(lambda) {}
  
  //! \name Generate samples
  //@{
  double number();
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return (1.0 / m_lambda);}
  double variance() const {return (1.0 / (m_lambda * m_lambda));}
  double stdDev() const {return (1.0 / m_lambda);}
  //@}

  //! \name Parameters of the random variable
  //@{
  /*! \return The parameter of the distribution. */
  double lambda() const {return m_lambda;}
  //@}

  //! \name Change parameters of the random variable
  //@{
  /*! Set the mean of the random variable.
    \param mean The new mean of the random variable. */
  void setMean(const double mean) {m_lambda = (1.0 / mean);}
  /*! Set the variance of the random variable.
    \param variance The new variance of the random variable. */
  void setVariance(const double variance) {m_lambda = sqrt(1.0 / variance);}
  /*! Set the standard deviation of the random variable.
    \param std_dev The new standard deviation of the random variable. */
  void setStdDev(const double std_dev) {m_lambda = (1.0 / std_dev);}
  /*! Set the parameter value of the distribution.
    \param lambda The new parameter of the distribution. */
  void setLambda(const double lambda) {m_lambda = lambda;}
  /*! Set the mean and variance of the random variable.
    \param mean The new mean of the random variable.
    \param variance The new variance of the random variable. */
  void setParameters(const double mean, const double variance) {
      setMean(mean); setVariance(variance);}
  //@}

private:
  double m_lambda;
};


//! Discrete uniform random number generator.
/*! This class generates discrete random numbers from a uniform distribution.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class DiscreteUniformRandom : public IntegerRandom {
public:
  /*! This constructor sets the low and high values for the distribution, 
    and the random number seed.
    \param low Low value for generated numbers.
    \param high High value for generated numbers.
    \param seed Random number seed.  Do \e not set this to zero.
    \note The numbers generated by this class are in the range low to high,
    inclusive.
  */
  explicit DiscreteUniformRandom(const int low, const int high,
		       const unsigned long seed = 1UL)
    :IntegerRandom(seed), m_low((low < high)?low:high),
     m_range((unsigned long)((low < high)?(high - low):(low - high))) {}
  /*! This constructor sets the high value for the distribution, and the
    random number seed.  The low value is assumed to be 0.
    \param range Range of discrete numbers to be generated.
    \param seed Random number seed.  Do \e not set this to zero.
    \note The numbers generated by this class are in the range 0 to
    (range - 1), inclusive.
  */
  explicit DiscreteUniformRandom(const int range, const unsigned long seed = 1UL)
    :IntegerRandom(seed), m_low(0), m_range(range - 1) {}

  //! \name Generate samples
  //@{
  int number() {return MTRand::randInt(m_range) + m_low;}
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return (((double)m_range / 2.0));}
  double variance() const {return ((double)(m_range * m_range) / 12.0);}
  double stdDev() const {return sqrt(variance());}
  //@}

  //! \name Parameters of the random variable
  //@{
  /*! \return The low value of the range of numbers that can be generated. */
  int lowValue() const {return m_low;}
  /*! \return The high value of the range of numbers that can be generated. */
  int highValue() const {return (m_low + m_range);}
  /*! \return The range of the numbers that can be generated. */
  int range() const {return m_range + 1;}
  //@}

  //! \name Change parameters of the random variable
  //@{
  /*! Set the low value for the random variable.
    \param low The new low value. */
  void setLowValue(const int low) {m_low = low;}
  /*! Set the high value for the random variable.
    \param high The new high value. */
  void setHighValue(const int high) {m_range = (unsigned long)(high - m_low);}
  /*! Set the both the low and high values of the random variable.
    \param low Low high value.
    \param high New high value. */
  void setValues(const int low, const int high) {
    setLowValue(low); setHighValue(high);}
  //@}

private:
  int m_low;
  unsigned long m_range;
};


//! Bernoulli random number generator.
/*! This class generates random numbers from a Bernoulli distribution.
  \author Bill Smart
  \date November 2001
  \version 1.0
  \ingroup gen_math_group math_group
*/
class BernoulliRandom : public IntegerRandom {
public:
  /*! This constructor sets the success probability, and the random number
    seed.
    \param success The probability of a successful trial.
    \param seed Random number seed.  Do \e not set this to zero.
  */
  explicit BernoulliRandom(const double success = 0.5, const unsigned long seed = 1UL)
    :IntegerRandom(seed), m_success(success) {}

  //! \name Generate samples
  //@{
  int number() {return (rand() <= m_success)?1:0;}
  //@}

  //! \name Moments of the random variable
  //@{
  double mean() const {return m_success;}
  double variance() const {return (m_success * (1.0 - m_success));}
  double stdDev() const {return sqrt(variance());}
  //@}

  //! \name Parameters of the random variable
  //@{
  /*! \return The probability of success. */
  double success() const {return m_success;}
  /*! \return The probability of success. */
  double probSuccess() const {return success();}
  /*! \return The probability of failure. */
  double failure() const {return (1.0 - m_success);}
  /*! \return The probability of failure. */
  double probFailure() const {return failure();}
  //@}

  //! \name Parameters of the random variable
  //@{
  /*! Set the probability of success for a trial.
    \param success New probability of success. */
  void setProbSuccess(const double success) {m_success = success;}
  /*! Set the probability of failure for a trial.
    \param failure New probability of failure. */
  void setProbFailure(const double failure) {m_success = 1.0 - failure;}
  //@}

private:
  double m_success;
};


// This is ugly, but half of the cost of these calls is in the function
// call overhead.


inline MTRand::uint32 MTRand::randInt() {
  if(left == 0)
    reload();
  --left;
		
  register uint32 s1;
  s1 = *pNext++;
  s1 ^= (s1 >> 11);
  s1 ^= (s1 <<  7) & 0x9d2c5680U;
  s1 ^= (s1 << 15) & 0xefc60000U;
  return (s1 ^ (s1 >> 18));
}


inline MTRand::uint32 MTRand::randInt(const uint32 &n) {
  // Find which bits are used in n
  uint32 used = ~0;
  for(uint32 m = n; m; used <<= 1, m >>= 1 ) {}
  used = ~used;
	
  // Draw numbers until one is found in [0,n]
  uint32 i;
  do
    i = randInt() & used;  // toss unused bits to shorten search
  while(i > n);
  return i;
}


inline void MTRand::randomize() {
#ifdef _MSC_VER
  unsigned long seed_number = (unsigned)time( NULL );
#else
  unsigned long seed_number;
  int fp = open("/dev/urandom", O_RDONLY);
  read(fp, &seed_number, sizeof(seed_number));
  close(fp);
#endif
  MTRand::randomize(seed_number);
}



inline void MTRand::randomize(unsigned long seed) {
  // Seed the generator with a simple uint32
  register uint32 *s;
  register int i;
  for(i = N, s = state; i--;
      *s = seed & 0xffff0000,
	*s++ |= ((seed *= 69069U)++ & 0xffff0000) >> 16,
	(seed *= 69069U)++) {}  // hard to read, but fast
  reload();
}


inline void MTRand::reload() {
  // Generate N new values in state
  // Made clearer and faster by Matthew Bellew (matthew.bellew@home.com)
  register uint32 *p = state;
  register int i;
  for(i = N - M; i--;)
    *p++ = twist(p[M], p[0], p[1]);
  for(i = M; --i;)
    *p++ = twist(p[M-N], p[0], p[1]);
  *p = twist(p[M-N], p[0], state[0]);
  left = N, pNext = state;
}


// We could cache another number m_std_dev * r * sin(theta) + m_mean, but
// this introduces non-randomness into the stream of numbers, although it
// speeds things up a bit.
inline double NormalRandom::number() {
  double theta = 2.0 * M_PI * MTRand::rand();
  double r = sqrt(-2.0 * log(MTRand::rand()));
  
  return ((m_std_dev * r * cos(theta)) + m_mean);
}


#endif
