#ifndef SoftStateAggregator_H
#define SoftStateAggregator_H

#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <iostream>
#include <vector>

#include <Random.H>
#include <Array.H>

#include <Approximator.H>
#include <Experience.H>

/// \brief Soft-state aggregation based Q-learning in continuous state space
/// and discrete action MDPs.
/// 
/// The idea is to allow 'overlapping' aggregated states.
/// The algorithm is described in 
/// Singh, S., Jaakkola, T., & Jordan, M. (1995). Reinforcement learning 
/// with soft state aggregation.NIPS 7 (pp. 361{368). MIT Press.
/// see also
/// http://citeseer.ist.psu.edu/82870.html
/// 
/// During training only a single Q-value is updated which
/// is selected at random from the ones competing for the given
/// training sample. 
class SoftStateAggregator : public Approximator {
public:
  SoftStateAggregator(const int dimension, const int actions, const int x_number,
	const double gain, const double discount, const double bandwidth);
  SoftStateAggregator(const int dimension, const int actions, const int x_number,
	const char *filename, const double gain, const double discount,
	const double bandwidth);

  void initialize(RealRandom &generator);

  virtual void learn(RnPoint &point, const int action, double value);
 
  void learn(RnPoint &state, const int action, const double reward,
	     RnPoint &next_state, bool goalState) {
    Experience e (state, action, reward, next_state,goalState);
  
    m_cache.push_back(e);
  }
  void actualLearn(RnPoint &state, const int action, const double reward,
		   RnPoint &next_state, bool goalState);
  void runCache();
  
  virtual double present(const RnPoint &point, const int action);
 
  virtual void read(std::istream &in);
  virtual void write(std::ostream &out) const;

  int numPoints() const {return m_x_number;}

private:
  int m_x_number;
  double m_gain;
  double m_discount;
  double m_bandwidth;

  Array2d <double> m_values;
  Array2d <int> m_counts;
  UniformRandomSimple random_chance;

  std::vector<RnPoint> m_data;
  std::vector<Experience> m_cache;

  double P(const RnPoint &point, const int n);

  int getRandomIndex(const RnPoint &point);
};


#endif

