#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#endif

#include <SoftStateAggregator.H>
#include <fstream>
#include <debug.h>
#include <Geometry.H>

using namespace std;


SoftStateAggregator::SoftStateAggregator(const int dimension, const int actions, const int x_number,
	     const double gain, const double discount, const double bandwidth)
  :Approximator(dimension, actions), m_x_number(x_number), m_gain(gain),
   m_discount(discount), m_bandwidth(bandwidth),
   m_values(x_number, actions, 0.0), m_counts(x_number, actions, 0) {
  assert(x_number > 0);
  assert_range_inc(gain, 0.0, 1.0);
  assert_range_inc(discount, 0.0, 1.0);
  assert(bandwidth > 0.0);

  random_chance.randomize();

  UniformRandomSimple random_state;
  RnPoint point(dimension);
  
  for(int i = 0; i < x_number; ++i) {
    for(int j = 0; j < dimension; ++j)
      point[j] = random_state;

    m_data.push_back(point);
  }
}



SoftStateAggregator::SoftStateAggregator(const int dimension, const int actions, const int x_number,
	     const char *filename, const double gain, const double discount,
	     const double bandwidth)
  :Approximator(dimension, actions), m_x_number(x_number), m_gain(gain),
   m_discount(discount), m_bandwidth(bandwidth),
   m_values(x_number, actions, 0.0), m_counts(x_number, actions, 0) {
  assert(x_number > 0);
  assert_range_inc(gain, 0.0, 1.0);
  assert_range_inc(discount, 0.0, 1.0);
  assert(bandwidth > 0.0);

  random_chance.randomize();

  RnPoint point(dimension);
  ifstream in(filename);
  for(int i = 0; i < x_number; ++i) {
    in >> point;
    m_data.push_back(point);
  }

  cout << "+++ " << m_data.size() << endl;
}


void SoftStateAggregator::initialize(RealRandom &generator) {
  for(int i = 0; i < m_x_number; ++i)
    for(int j = 0; j < numActions(); ++j) {
      m_values[i][j] = generator;
      m_counts[i][j] = 0;
    }
}



void SoftStateAggregator::learn(RnPoint &point, const int action, const double value) {
  cout << "learn" << endl;
  abort();
}



void SoftStateAggregator::actualLearn(RnPoint &state, const int action,
			const double reward, RnPoint &next_state,
			bool goalState) {
  assert_range(action, 0, numActions());

  const int x_state = getRandomIndex(state);
  
  const double local_gain = m_gain / (double)
    (1 + (m_counts[x_state][action] / 1000));

  const int next_x_state = getRandomIndex(next_state);
  int max_q_action = 0;
  for(int i = 1; i < numActions(); ++i)
    if (m_values[next_x_state][i] > m_values[next_x_state][max_q_action])
      max_q_action = i;
  
  const double backup_value = (goalState)?0:
    m_discount * m_values[next_x_state][max_q_action];
  m_values[x_state][action] = (1.0 - local_gain) * m_values(x_state, action) +
    local_gain * (reward + backup_value);
  
  ++m_counts[x_state][action];
  
  m_flops += 5;
}



void SoftStateAggregator::runCache() {
  vector<Experience>::reverse_iterator it;
  for(it = m_cache.rbegin(); it != m_cache.rend(); ++it) {
    actualLearn(it->state, it->action, it->reward, it->next_state, it->goalState);
  }
 
  while (!m_cache.empty())
    m_cache.pop_back();
}



double SoftStateAggregator::present(const RnPoint &point, const int action) {
  assert_range(action, 0, numActions());

  double q = 0.0;
  for(int i = 0; i < m_x_number; ++i)
    q += P(point, i) * m_values(i, action);

  return q;
}



void SoftStateAggregator::read(std::istream &in) {
  // Empty the list of points
  while (!m_data.empty())
    m_data.pop_back();
 
  RnPoint point(dimension());
 
  in >> point;
  while (!in.eof()) {
    m_data.push_back(point);
     
    in >> point;
  }

  m_x_number = m_data.size();
}



void SoftStateAggregator::write(std::ostream &out) const {
  for(vector<RnPoint>::const_iterator it = m_data.begin();
      it != m_data.end(); ++it)
    out << *it << endl;
}


// Return P_n(point)
double SoftStateAggregator::P(const RnPoint &point, const int n) {
  // Calculate the total kernel
  double kernel_total = 0.0;
  for(vector<RnPoint>::const_iterator it = m_data.begin();
      it != m_data.end(); ++it)
    kernel_total += kernel(it->distance(point), m_bandwidth);

  return kernel((m_data.begin() + n)->distance(point), m_bandwidth) /
    kernel_total;
}



int SoftStateAggregator::getRandomIndex(const RnPoint &point) {
  const double random_number = random_chance;

  double cumulative = 0.0;
  for(int i = 0; i < m_x_number; ++i) {
    cumulative += P(point, i);
    if (random_number <= cumulative)
      return i;
  }

  cout << random_number << ' ' << cumulative << endl;
  abort();
}

