#ifdef _MSC_VER
# pragma warning(disable:4786)
# pragma warning(disable:4503) // warning: decorated name length exceeded
#include <process.h>
#else
#include <unistd.h>
#endif

#include <iostream>
#include <cstdlib>
#include <sstream>
#include <iomanip>
#include <cstdio>

#include <sys/types.h>
#include <memory>

#include <Approximator.H>
#include <HillCar.H>
#include <HardStateAggregator.H>
#include <Random.H>
#include <Geometry.H>
#include <DescriptiveStats.H>

#include <IFAPPQ.H>
#include <KBRL.H>
#include <SoftStateAggregator.H>

using namespace std;

/// Converts a value (typically numeric) to a string representation.
template <class T> std::string toString( T value, int width = -1 )
{
  std::ostringstream stream;
  if (width>0)
    stream << std::setw( width ) << std::setfill( '0' );
  stream << value;
  return stream.str();
}

ofstream results;
ofstream os_mse;

// The system is 'car-on-the-hill' for the time being.
HillCar car;
// Function approx
std::auto_ptr<Approximator> approx;

const int TABLE_TRAINING_STEPS = 1000000;
const int TABLE_TEST_SAMPLES  = 100;

const int TRAINING_RUNS = 110;
const int EVAL_EVERY = 10;
const int RUN_CACHE_EVERY = 1;

const int MAX_STEPS = 1000;

const int DISCRETE_STATES = 30; // tabular Q-learning table size
const int MEAN_GROUP_SIZE = 1;
double ADAPTIVE_THRESHOLD = 0.1;

const double TABLE_GAIN = 0.2;
double GAIN = 1.0;
const double DISCOUNT = 0.99;
const double INITIAL_VALUE = 0.0;
const double EPSILON = 0.2;
const double COUNT_FACTOR = 1000;

const int SINGH_NUMBER = 200;
const double SINGH_BANDWIDTH = 0.1;

const bool RANDOM_PLACEMENT = true;

const double VALUE_ITERATION_EPS = 0.01;
const int    VALUE_ITERATION_ITMAX = 100000;
const double KBRL_PROB_FLOOR = 0.0000001;

const double GRAPH_STEPS = 30.0;

const int STATE_ACTION_TOTALS_BOUND = 20;

void parseArguments(int argc, char **argv);
void openLogFile(const char *const tag);

void dumpValues( HillCar& car, Approximator& approx,
		 const std::string& fname, const double inc);

int main(int argc, char **argv) {
  parseArguments(argc, argv);

  UniformRandomSimple random_state;
  UniformRandom random_value(-0.01, 0.01);
  DiscreteUniformRandom random_action(car.numActions());
  DiscreteUniformRandom random_action_add(car.numActions() - 1);
  UniformRandomSimple random_chance;


  //approx->initialize(random_value);
  MDPStateDiscretizer discrete(car, DISCRETE_STATES);

  // How many states should we have in the Q-table to make things fair?
  int table_states = 1;
  for(int i = 0; i < car.stateDimension(); ++i)
    table_states *= DISCRETE_STATES;

  RnPoint point(car.stateDimension());
  const double inc = 1.0 / GRAPH_STEPS;
    
  HardStateAggregator table(table_states, car.numActions(), TABLE_GAIN, DISCOUNT);

  // Make sure we have a properly trained table
  int steps = 0;
  int avg_steps = 0;
  int vsteps = 0;
  do {
    table.initialize(random_value);

    // Train the table
    results << "Training HardStateAggregator: " << flush;
    for(int i = 0; i < TABLE_TRAINING_STEPS; ++i) {
      
      // draw a non-goal state
      do
      {
        for(int j = 0; j < point.n(); ++j)
	        point[j] = random_state;
        car.setNormalizedState(point);
      } while (car.goalState());

      // draw an action
      int action;

      // Hack to make two actions work
      do {
        action = random_action;
      } while (action == 1);

      int state = discrete.state();
      double reward = car.applyAction(action);
      int next_state = discrete.state();      
      table.learn(state, action, reward, next_state, car.goalState());
    }
    results << "done" << endl;

    // Test the table
    steps = 0;
    avg_steps = 0;
    vsteps = 0;
    for (int testi = 0; testi<TABLE_TEST_SAMPLES; ++testi) {
      int tsteps = 0;
      if (testi==0)
        car.reset();
      else
        car.setRandomState();
      while ((tsteps < MAX_STEPS) && (!car.goalState())) {
      	// Two actions hack
        if (table.q(discrete.state(), 0) > table.q(discrete.state(), 2))
	        car.applyAction(0);
        else
	        car.applyAction(2);

        ++tsteps;
      }
      avg_steps += tsteps;
      steps = (tsteps>steps?tsteps:steps);
      if (testi==0)
        vsteps = tsteps;
    }
    results << "steps from middle:" << vsteps << endl;
    results << "avg steps :" << avg_steps/TABLE_TEST_SAMPLES << endl;
    results << "worst steps :" << steps << endl;
  } //while (steps == 1000);
  while (steps > 148);

  results << "HardStateAggregator performance: " << vsteps << " steps" << endl << endl;

  // Draw (some of) the table(s)
  {
    vector<vector<double> > tbl0(DISCRETE_STATES,vector<double>(DISCRETE_STATES,0.0) );
    vector<vector<double> > tbl2(DISCRETE_STATES,vector<double>(DISCRETE_STATES,0.0) );
    {
      int count = 0;
      for(int i = 0; i < DISCRETE_STATES; ++i) 
        for(int j = 0; j < DISCRETE_STATES; ++j, ++count) {
          tbl0[j][i] = table.q(count,0);
          tbl2[j][i] = table.q(count,2);
        }
    }
    {
      ofstream t0("table.0"), t2("table.2"), td("table.02.diff");
      for(int i = 0; i < DISCRETE_STATES; ++i) {
        for(int j = 0; j < DISCRETE_STATES; ++j) {
          t0 << tbl0[i][j] << endl;
          t2 << tbl2[i][j] << endl;
          td << tbl0[i][j]-tbl2[i][j] << endl;
        }
        t0 << endl;
        t2 << endl;
        td << endl;
      }
    }
  }

  Array2d <int> state_action_hits(discrete.numStates(),
				  car.numActions(), 0);
  Array2d <int> state_action_totals(discrete.numStates(),
				    car.numActions(), 0);

  Array2d <int> state_action_totals_fixed(discrete.numStates(),
				    car.numActions(), 0);

  {
    std::ifstream is_vc0( "i_visit_cnts_0_05000.dat" );
    std::ifstream is_vc2( "i_visit_cnts_2_05000.dat" );
    assert( is_vc0.good() );
    assert( is_vc2.good() );
    RnPoint state(car.stateDimension());
    for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
      for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
      	car.setNormalizedState(state);
        int c0, c2;
        is_vc0 >> c0;
        is_vc2 >> c2;
	      state_action_totals_fixed[discrete.state()][0] = c0;
	      state_action_totals_fixed[discrete.state()][2] = c2;
      }
    }
  }

  // Train the function approximator
  RnPoint state(car.stateDimension());
  RnPoint next_state(car.stateDimension());
  int total_steps = 0;
  {
    for(int i = 0; i < TRAINING_RUNS; ++i) {
      state_action_hits = 0;

      // Do one training run
      int steps = 0;

      //    car.reset();
      car.setRandomState();
      while ((steps < MAX_STEPS) && (!car.goalState())) {
      	// Two action hack
        int action;
        if (table.q(discrete.state(), 0) > table.q(discrete.state(), 2))
	        action = 0;
        else
	        action = 2;

        // Epsilon?
        if (random_chance < EPSILON)
        {
	        // Two action hack
	        do {
	          action = (action + random_action_add + 1) % (car.numActions());
	        } while (action == 1);
        }

        ++state_action_hits[discrete.state()][action];// changed 21/04/2004 13:50CET, Cs.Sz

        car.normalizedState(state);
        double reward = car.applyAction(action);
        car.normalizedState(next_state);

        approx->learn(state, action, reward, next_state,car.goalState());

        ++steps;
      }
      total_steps += steps;

      if ((i % RUN_CACHE_EVERY) == (RUN_CACHE_EVERY-1))
        approx->runCache();

      for(int j = 0; j < discrete.numStates(); ++j)
        for(int k = 0; k < car.numActions(); ++k)
          state_action_totals[j][k] += state_action_hits[j][k];

      if ((i % EVAL_EVERY) == (EVAL_EVERY - 1)) {
        // Run a test
        steps = 0;
        car.reset();
        while ((steps < MAX_STEPS) && (!car.goalState())) {
	        car.normalizedState(state);

      		// Two action hack
	        if (approx->q(state, 0) > approx->q(state, 2))
	          car.applyAction(0);
	        else
	          car.applyAction(2);

	        ++steps;
        }

        // Get the MSE of the differences
        DescriptiveStats mse;
        mse.reset();
        for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
	        for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
      	    car.setNormalizedState(state);

	          // Two actions
	          for(int k = 0; k < car.numActions(); k += 2) {
	            if (state_action_totals_fixed[discrete.state()][k] >
    		            STATE_ACTION_TOTALS_BOUND) {
      		      double a = table.q(discrete.state(), k);
      		      double b = approx->q(state, k);
      		      mse += (a - b) * (a - b); 
              }
	          }
	        }
        }

        results << i + 1 << '\t' << sqrt(mse.mean()) << " # mean" << endl <<
	      i + 1 << '\t' << approx->numPoints() << " # points" << endl <<
	      i + 1 << '\t' << steps << " # steps" << endl;
	  
        char flop_buffer[1024];
        sprintf( flop_buffer, "%lli", approx->flops() );
        os_mse << i+1 << "\t" 
               << flop_buffer << "\t" 
               << sqrt(mse.mean()) << "\t"
               << approx->numPoints() << "\t"
               << approx->mem_usage() << "\t"
               << steps << "\t"
               << endl;

        const int action_inc = 2;
    	  for(int k = 0; k < car.numActions(); k += action_inc) {
          std::ofstream os_vp( (std::string("visit_probs_")
				                        +toString(k)+"_"
				                        +toString(i+1,5)+".dat").c_str() );
          std::ofstream os_vc( (std::string("visit_cnts_")
				                        +toString(k)+"_"
				                        +toString(i+1,5)+".dat").c_str() );
          for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
  	        for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
	            car.setNormalizedState(state);
    	        os_vp 
          		  << ((double)state_action_totals[discrete.state()][k] / (double)total_steps) 
                << "\t";
	            os_vc 
                << state_action_totals[discrete.state()][k] << "\t";
	          }
            os_vp << "\n";
            os_vc << "\n";
          }
        }
        dumpValues( car, *approx, 
		    std::string("value.")+toString(i+1,5), inc );
      } // EVAL_EVERY
    } // TRAINING_RUNS
  }
  
  // Draw the policy
#ifdef _MSC_VER
  vector<ofstream> out(car.numActions());
#else
  ofstream out[car.numActions()];
#endif
  {
    for(int i = 0; i < car.numActions(); ++i) {
      char buffer[256];
      sprintf(buffer, "policy.%i", i);
      out[i].open(buffer);
    }
  }
  
  for(state[1] = inc / 10.0; state[1] <= 1.0; state[1] += inc / 5.0)
    for(state[0] = inc / 10.0; state[0] <= 1.0; state[0] += inc / 5.0)
      // Two action hack
      if (approx->q(state, 0) > approx->q(state, 2))
      	out[0] << state << endl;
      else
      	out[2] << state << endl;

  {
    for(int i = 0; i < car.numActions(); ++i)
      out[i].close();
  }

  // Draw the values
  dumpValues( car, *approx, "value", inc );
  {
    ofstream learned("approx");
    approx->write(learned);
  }

  return 0;
}


void dumpValues( HillCar& car, Approximator& approx,
		 const std::string& fname 
   , const double inc
   )
{
  RnPoint state(car.stateDimension());
  for(int i = 0; i < car.numActions(); ++i)
  {
    std::ofstream out( (fname+"."+toString(i)).c_str() );
    for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
      for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
      	out << approx.q(state, i) << endl;
      }
      out << endl;
    }
  }

  {
    std::ofstream out( (fname+".policy").c_str() );
    for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
      for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
        if (approx.q(state, 0)<approx.q(state, 2))
          out << "0";
        else
          out << "2";
        out << endl;
      }
      out << endl;
    }
  }

  {
    std::ofstream out( (fname+".diff").c_str() );
    for(state[1] = inc / 2.0; state[1] <= 1.0; state[1] += inc) {
      for(state[0] = inc / 2.0; state[0] <= 1.0; state[0] += inc) {
        out << (approx.q(state, 0)-approx.q(state, 2)) << endl;
      }
      out << endl;
    }
  }
}

void parseArguments(int argc, char **argv) {
  if (argc != 4) {
    cerr << "Usage:" << endl << "  " << argv[0] << " <threshold> <gain> " <<
      "{-rlfapp | -kbrl | -ssa}" << endl;
    exit(1);
  }

  ADAPTIVE_THRESHOLD = atof(argv[1]);
  GAIN = atof(argv[2]);

  char tag[256];
  if (!strcmp("-rlfapp", argv[3])) {
    approx.reset( new IFAPPQ(car.stateDimension(), car.numActions(),
			GAIN, DISCOUNT, ADAPTIVE_THRESHOLD,
			0.01, COUNT_FACTOR, INITIAL_VALUE,
			RANDOM_PLACEMENT) );
    sprintf(tag, "rlfapp");
  } else if (!strcmp("-kbrl", argv[3])) {
    approx.reset( new KBRL(car.stateDimension(), car.numActions(),
		      DISCOUNT, VALUE_ITERATION_EPS,
		      VALUE_ITERATION_ITMAX, KBRL_PROB_FLOOR) );
    sprintf(tag, "kbrl");
  } else if (!strcmp("-ssa", argv[3])) {
    approx.reset( new SoftStateAggregator(car.stateDimension(), car.numActions(), 264,
		       "ssa-points", GAIN, DISCOUNT, SINGH_BANDWIDTH) );
    sprintf(tag, "ssa");
  } else {
    cerr << "Unknown approximation scheme: " << argv[3] << endl;
    exit(1);
  }
  openLogFile(tag);
}

void openLogFile(const char *const tag) {
  char buffer[256];
  sprintf(buffer, "results/%s-%i-t-%1.2f-g-%1.2f", tag, getpid(),
	  ADAPTIVE_THRESHOLD, GAIN);
  char buffer2[256];
  sprintf(buffer2, "results/%s-mse-%i-t-%1.2f-g-%1.2f", tag, getpid(),
	  ADAPTIVE_THRESHOLD, GAIN);

  results.open(buffer);
  os_mse.open(buffer2);
}

